;
; < ------------------------------- >
;   PureBasic IDE - StructureViewer 
; < ------------------------------- >
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version. See the License.txt file for more details.
; 
; (c) 2002 Fantaisie Software
;


XIncludeFile "Common.pb"

Enumeration #GADGET_StructureViewer
  #GADGET_StructureViewer_Panel
  #GADGET_StructureViewer_List
  #GADGET_StructureViewer_Parent
  #GADGET_StructureViewer_Name
  #GADGET_StructureViewer_Ok
  #GADGET_StructureViewer_Cancel
;  #GADGET_StructureViewer_OnTop
  #GADGET_StructureViewer_InsertName
  #GADGET_StructureViewer_InsertStruct
  #GADGET_StructureViewer_InsertCopy
  #GADGET_StructureViewer_InsertType
EndEnumeration

#SVM_Structures = 0  ; Associated to 'StructureViewerMode' variable, to know if we are in Structure or Interface mode
#SVM_Interfaces = 1

Global  StructureIndex, StructureViewerInSearch, StructureViewerMode

Structure StructHistory
  Name$
  Line.l
EndStructure

; How it works: we have 3 linked lists
; 1 which hold the data of the current displayed stuffs (Structure or Interfaces)
; 1 which hold the structure data
; 1 which hold the interface data
;
; It allows to use common functions to manage the list content and switch from one ot another list quickly.
;
NewList StructureList$()
NewList StructureListBuffer$()
NewList InterfaceListBuffer$()

NewList StructureHistory.StructHistory()
NewList InsertStructureItem$()

Procedure AddSortedStructure(StructureName$)

  StructureUName$ = UCase(StructureName$)
  
  ResetList(StructureList$())
  While Quit = 0

    If NextElement(StructureList$())
      If UCase(StructureList$()) >= StructureUName$
        Quit = 1
                
        If PreviousElement(StructureList$())
        Else
          ResetList(StructureList$())
        EndIf
      EndIf
    Else
      Quit = 1
    EndIf
  Wend
  
  AddElement(StructureList$())
  
  StructureList$() = StructureName$
EndProcedure


Procedure DisplayStructuresList()
  SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 0, 0)
  ClearGadgetItemList(#GADGET_StructureViewer_List)

  ForEach StructureList$()
    AddGadgetItem(#GADGET_StructureViewer_List, -1, StructureList$())
  Next
  
  SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 1, 0)
  InvalidateRect_(GadgetID(#GADGET_StructureViewer_List),0,0)
  DisableGadget(#GADGET_StructureViewer_InsertStruct, 1)
  DisableGadget(#GADGET_StructureViewer_InsertCopy  , 1)
  
  ClearList(StructureHistory()) ; Reset the history
EndProcedure


Procedure DisplayCorrectStructureList()

  If StructureViewerMode <> GetGadgetState(#GADGET_StructureViewer_Panel)  ; Protect it against several click on the same panel
        
    ForEach StructureList$() ; To fix the memory leak bug in PureBasic. To remove once it's fixed
      StructureList$() = ""
    Next
      
    ClearList(StructureList$())
    
    If GetGadgetState(#GADGET_StructureViewer_Panel) = 0  ; Structures

      ForEach StructureListBuffer$()
        AddElement(StructureList$())
        StructureList$() = StructureListBuffer$()
      Next

      StructureViewerMode = #SVM_Structures
      
    Else ; Interfaces

      ForEach InterfaceListBuffer$()
        AddElement(StructureList$())
        StructureList$() = InterfaceListBuffer$()
      Next
      
      StructureViewerMode = #SVM_Interfaces
    EndIf
    
    DisplayStructuresList()
  EndIf

EndProcedure


Procedure IsStructure(StructureName$)
  Result = -1
  
  ; Look in the 2 buffers to see if the word is a structure or an interface
  ;
  ForEach StructureListBuffer$()
    If UCase(StructureListBuffer$()) = UCase(StructureName$)
      Result = ListIndex(StructureListBuffer$())
      SetGadgetState(#GADGET_StructureViewer_Panel, 0)  ; Select the 'Structure' panel
    EndIf
  Next

  ForEach InterfaceListBuffer$()
    If UCase(InterfaceListBuffer$()) = UCase(StructureName$)
      Result = ListIndex(InterfaceListBuffer$())
      SetGadgetState(#GADGET_StructureViewer_Panel, 1)  ; Select the 'Interface' panel
    EndIf
  Next
  
  DisplayCorrectStructureList()  ; Display the right list (and change the current buffer to the right one)
  
  ProcedureReturn Result
EndProcedure


Procedure ResizeStructureViewerWindow()

  UseWindow(#WINDOW_StructureViewer)
  Width  = WindowWidth()
  Height = WindowHeight()
  Top    = 6
  x = (Width-116-17)/2

  ResizeGadget(#GADGET_StructureViewer_Panel       ,   -1,  -1, Width-12, -1) : Top + GadgetHeight(#GADGET_StructureViewer_Panel)
  ResizeGadget(#GADGET_StructureViewer_List        ,   -1,  -1, Width-12, Height-95-20) : Top + GadgetHeight(#GADGET_StructureViewer_List)+10
  ResizeGadget(#GADGET_StructureViewer_Parent      ,   -1, Top, -1, -1)
  ResizeGadget(#GADGET_StructureViewer_Name        ,   -1, Top, Width-GadgetWidth(#GADGET_StructureViewer_Parent)-18, -1) : Top+25
  ; ResizeGadget(#GADGET_StructureViewer_OnTop       ,   -1, Top, -1, -1)
  ResizeGadget(#GADGET_StructureViewer_InsertName  ,   -1, Top, x, -1)
  ResizeGadget(#GADGET_StructureViewer_InsertCopy  ,126+x, Top, Width-17-(116+x), -1) : Top+28
  ResizeGadget(#GADGET_StructureViewer_InsertStruct,126+x, Top, Width-17-(116+x), -1)
  ResizeGadget(#GADGET_StructureViewer_InsertType  ,   -1, Top, -1, -1)
EndProcedure


Procedure SearchStructure(Search$)

  Search$ = UCase(Search$)
  
  SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 0, 0)

  DisplayStructuresList()

  Index = 0

  If ValidCharacters(Asc(Search$))

    ResetList(StructureList$())

    CharToCheck  = 1
    CharToSearch = Asc(Search$) ;Asc(Mid(Search$,CharToCheck,1))

    ; Search Start of structures alphabetically
    While Found = 0 And NextElement(StructureList$())
      If CharToSearch > Asc(StructureList$())
        Index = ListIndex(StructureList$())
      ElseIf CharToSearch < Asc(StructureList$())
        Found = 2
      ElseIf CharToSearch = Asc(StructureList$())
        Index = ListIndex(StructureList$())
        Found = 1
      EndIf
    Wend

    ; Search End of structures alphabetically
    If Found = 1
      StopIndex = Index
      While Found = 1 And NextElement(StructureList$())
        If CharToSearch = Asc(StructureList$())
          StopIndex = ListIndex(StructureList$())
        ElseIf CharToSearch < Asc(StructureList$())
            Found = 2
        EndIf
      Wend

      ; OK, found Start and End
      If Found = 2 And (StopIndex <> Index) And ListIndex(StructureList$())

;        DEBUG TEST:
;         For x = Index To StopIndex
;           SelectElement(StructureList$(),x-1)
;           A$ + StructureList$()+Chr(13)
;         Next x
;         MessageRequester("INFO",A$,0)

        Found = 0
        For a = Len(Search$) To 1 Step -1
          A$ = Left(Search$,a)
          For b = StopIndex To Index Step -1
            SelectElement(StructureList$(),b)    ; LISTINDEX CHANGE ->  b-1
            If FindString(UCase(StructureList$()),A$,1)=1
              Found = b
            EndIf
          Next b
          If Found
            a = 1  ; BREAK
          EndIf
        Next a
        
        If Found
          Index = Found
        EndIf

      EndIf
    EndIf

  EndIf
  
  SetGadgetState(#GADGET_StructureViewer_List,Index)
  SetWindowText_(WindowID(#WINDOW_StructureViewer),Langage(112)+"  -  ."+Trim(GetGadgetText(#GADGET_StructureViewer_List)))

  SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 1, 0)
  InvalidateRect_(GadgetID(#GADGET_StructureViewer_List),0,0)

  DisableGadget(#GADGET_StructureViewer_InsertName, 0)
  ActivateGadget(#GADGET_StructureViewer_Name)

EndProcedure


Procedure SendCompilerRequest(RequestID)

  While WindowEvent() : Wend ; Ensure the message queue is empty to avoid loosing messages
  
  PostThreadMessage_(CompilerThreadID, PB_MSG_ID, RequestID, 0)

  ; Just wait for the compiler response with timeout.
  ; If compiler crashes, the timeout quits the procedure
  ; and the user can save his sources and nothing gets lost.
  ;
  Repeat
    
    EventID = WindowEvent()
  
    If EventID ; Don't go in the 'Delay' case if there is an event to handle
    
      If EventID = PB_MSG_ID And EventwParam() = RequestID
        ProcedureReturn 1
      EndIf
    Else
      Delay(20)
      Timer + 20
      If Timer >= 5000 ; ms
        MessageRequester("FATAL ERROR","PUREBASIC COMPILER CRASHED !!"+Chr(13)+Chr(13)+"Save all SourceCodes immediately and restart PureBasic !!",#MB_ICONERROR)
        Quit = 1
      EndIf
    EndIf
  Until Quit = 1

EndProcedure


Procedure.s GetStructMembers(Struct$)
  If CreateFile(#CREATEFILE_STRUCTUREINFOREQUEST,SharedFileName)
    WriteString(Struct$)
    CloseFile(#CREATEFILE_STRUCTUREINFOREQUEST)

    If SendCompilerRequest(#PB_MSG_Structure_Info)

      If ReadFile(#READFILE_STRUCTUREINFOREQUEST, FullPath + "Compilers\PBStructures.txt")
        Repeat
          Line$ = ReadString()
          If Line$
            member$+Trim(Line$)+"|"
          EndIf
        Until Line$ = ""
        CloseFile(#READFILE_STRUCTUREINFOREQUEST)
      EndIf
      
    EndIf
  EndIf
  
  ProcedureReturn member$
EndProcedure


Procedure AddExpandedStructureElement(String$,Prefix$)  ; Recursive, calls itself!
  A$ = UCase(Right(String$,2))
  If A$ = ".B" Or A$ = ".W" Or A$ = ".L" Or A$ = ".F" Or A$ = ".S" Or Right(String$,1) = "$"
    If StructureViewerIncludeType=0 And Right(A$,1)<>"$"
      String$ = Left(String$,Len(String$)-2)
    EndIf
    AddElement(InsertStructureItem$())
    InsertStructureItem$() = Prefix$+"\"+String$
  Else
    pos = FindString(String$,"[",1)
    If pos
      count = Val(Mid(String$,pos+1,FindString(String$,"]",1)-pos-1))-1
      len   = Len(Str(count))
      x     = FindString(String$,".",1)
      A$ = UCase(Mid(String$,x,pos-x))
      If A$ = ".B" Or A$ = ".W" Or A$ = ".L" Or A$ = ".F" Or A$ = ".S" Or Mid(String$,pos-1,1) = "$"
        A$ = Left(String$,pos-1)
        If StructureViewerIncludeType=0 And Right(A$,1)<>"$"
          A$ = Left(A$,Len(A$)-2)
        EndIf
        If count <= 9
          For a = 0 To count
            AddElement(InsertStructureItem$())
            InsertStructureItem$() = Prefix$+"\"+A$+"["+RSet(Str(a),len)+"]"
          Next a
        Else
          AddElement(InsertStructureItem$()) :  InsertStructureItem$() = ""
          For a = 0 To 2
            AddElement(InsertStructureItem$())
            InsertStructureItem$() = Prefix$+"\"+A$+"["+RSet(Str(a),len)+"]"
          Next a
          AddElement(InsertStructureItem$())
          InsertStructureItem$() = Prefix$+"\"+A$+"["+RSet(".",len,".")+"]"
          AddElement(InsertStructureItem$())
          InsertStructureItem$() = Prefix$+"\"+A$+"["+RSet(Str(count),len)+"]"
          AddElement(InsertStructureItem$()) :  InsertStructureItem$() = ""
        EndIf
      Else
        Name$    = Left(String$,x-1)
        Struct$  = Mid(String$,x+1,pos-x-1)
        If IsStructure(Struct$) <> -1
          Members$ = GetStructMembers(Struct$)
          If count <= 3
            For a = 0 To count
              x = 1
              Repeat
                B$ = StringField(Members$,x,"|"):x+1
                If B$
                  AddExpandedStructureElement(B$,Prefix$+"\"+Name$+"["+RSet(Str(a),len)+"]")
                EndIf
              Until B$ = ""
            Next a
          Else
            AddElement(InsertStructureItem$()) :  InsertStructureItem$() = ""
            For a = 0 To 1
              x = 1
              Repeat
                B$ = StringField(Members$,x,"|"):x+1
                If B$
                  AddExpandedStructureElement(B$,Prefix$+"\"+Name$+"["+RSet(Str(a),len)+"]")
                EndIf
              Until B$ = ""
            Next a
            x = 1
            Repeat
              B$ = StringField(Members$,x,"|"):x+1
              If B$
                AddExpandedStructureElement(B$,Prefix$+"\"+Name$+"["+RSet(".",len,".")+"]")
              EndIf
            Until B$ = ""
            x = 1
            Repeat
              B$ = StringField(Members$,x,"|"):x+1
              If B$
                AddExpandedStructureElement(B$,Prefix$+"\"+Name$+"["+RSet(Str(count),len)+"]")
              EndIf
            Until B$ = ""
            AddElement(InsertStructureItem$()) :  InsertStructureItem$() = ""
          EndIf
        Else
          AddElement(InsertStructureItem$())
          InsertStructureItem$() = Prefix$+"\"+String$
        EndIf

      EndIf
    Else
      pos = FindString(String$,".",1)
      Name$    = Left(String$,pos-1)
      Struct$  = Mid(String$,pos+1,Len(String$))
      If IsStructure(Struct$) <> -1
        Members$ = GetStructMembers(Struct$)
        x = 1
        Repeat
          B$ = StringField(Members$,x,"|"):x+1
          If B$
            AddExpandedStructureElement(B$,Prefix$+"\"+Name$)
          EndIf
        Until B$ = ""
      Else
        AddElement(InsertStructureItem$())
        InsertStructureItem$() = Prefix$+"\"+String$
      EndIf
    EndIf
  EndIf
EndProcedure


Procedure InsertStructure()
  SendMessage_(*RichEditGadget, #EM_REPLACESEL   , 1, "")
  SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 1, 0)
  SendMessage_(*RichEditGadget,#EM_EXGETSEL,0,Range.CHARRANGE)
  line   = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,Range\cpMin)
  Offset = Range\cpMin-SendMessage_(*RichEditGadget,#EM_LINEINDEX,line,0)

  ClearList(InsertStructureItem$())
  For a = 1 To CountGadgetItems(#GADGET_StructureViewer_List)-2
    AddExpandedStructureElement(Trim(GetGadgetItemText(#GADGET_StructureViewer_List,a,0)),"")
  Next a

  a = 0
  ResetList(InsertStructureItem$())
  While NextElement(InsertStructureItem$())
    x = Len(InsertStructureItem$())
    If x > a
      a = x
    EndIf
  Wend
  a + 1

  variable$ = InputRequester("Get variable name","Please input variable name","var")

  ResetList(InsertStructureItem$())
  While NextElement(InsertStructureItem$())
    If InsertStructureItem$()
      InsertStructureItem$() = Space(Offset)+Trim(variable$)+LSet(InsertStructureItem$(),a)+"= "
    EndIf
  Wend

  RichEdit_InsertString(variable$+"."+StructureHistory()\Name$+Chr(10), #False)
  ResetList(InsertStructureItem$())
  While NextElement(InsertStructureItem$())
    RichEdit_InsertString(InsertStructureItem$()+Chr(10), #False)
  Wend
  
  UpdateLineNumbering()
  UpdateProcedureList()

  ; set Cursor after first member
  index  = SendMessage_(*RichEditGadget,#EM_LINEINDEX,line+1,0)
  Range\cpMin = index + SendMessage_(*RichEditGadget,#EM_LINELENGTH,index,0)
  Range\cpMax = Range\cpMin
  SendMessage_(*RichEditGadget,#EM_EXSETSEL,0,Range.CHARRANGE)

EndProcedure


Procedure DisplayStructureInfo(StructureName$, oldIndex, AddHistory)

  If CreateFile(#CREATEFILE_STRUCTUREINFOREQUEST, SharedFileName)
    WriteString(StructureName$)
    CloseFile(#CREATEFILE_STRUCTUREINFOREQUEST)
    
    If StructureViewerMode = #SVM_Structures  ; Structures
    
      SendCompilerRequest(#PB_MSG_Structure_Info)
      
      BlockStart$ = "Structure "+UCase(StructureName$)
      BlockEnd$   = "EndStructure"
      Filename$   = "Compilers\PBStructures.txt"
      
    Else ; Interfaces

      SendCompilerRequest(#PB_MSG_Interface_Info)
      
      BlockStart$ = "Interface "+StructureName$
      BlockEnd$   = "EndInterface"    
      Filename$   = "Compilers\PBInterfaces.txt"
    
    EndIf
    
    SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 0, 0)

    ClearGadgetItemList(#GADGET_StructureViewer_List)
    AddGadgetItem(#GADGET_StructureViewer_List, -1, BlockStart$)
  
    If ReadFile(#READFILE_STRUCTUREINFOREQUEST, FullPath + Filename$)
      Repeat
        Line$ = ReadString()
        Line$ = ReplaceString(Line$,"$.s","$",1) ; remove .s on $
        If Line$
          AddGadgetItem(#GADGET_StructureViewer_List, -1, "  "+Line$)
        EndIf
      Until Line$ = ""
      CloseFile(#READFILE_STRUCTUREINFOREQUEST)
    EndIf

    AddGadgetItem(#GADGET_StructureViewer_List, -1, BlockEnd$)
    SetGadgetText(#GADGET_StructureViewer_Name, StructureName$)
    SetGadgetState(#GADGET_StructureViewer_List, 0)
    SetWindowText_(WindowID(#WINDOW_StructureViewer),Langage(112)+"  -  ."+StructureName$)
    
    If StructureViewerMode = #SVM_Structures
      DisableGadget(#GADGET_StructureViewer_InsertName, 0)
      DisableGadget(#GADGET_StructureViewer_InsertStruct, 0)
      DisableGadget(#GADGET_StructureViewer_InsertCopy  , 0)
    EndIf

    SendMessage_(GadgetID(#GADGET_StructureViewer_List), #WM_SETREDRAW, 1, 0)
    InvalidateRect_(GadgetID(#GADGET_StructureViewer_List),0,0)

    ActivateGadget(#GADGET_StructureViewer_List)
    Quit = 1
    
    If AddHistory    
      AddElement(StructureHistory())
      StructureHistory()\Name$ = StructureName$
      StructureHistory()\Line  = oldIndex
    EndIf

    If CountList(StructureHistory()) = 1
      DisableGadget(#GADGET_StructureViewer_Parent, 0)
    EndIf
  EndIf
EndProcedure


Procedure OpenStructureViewerWindow()

  CurrentWord$   = GetCurrentWord()

  If IsStructureViewerWindow = 0

    ClearList(StructureHistory())
    
    If OpenWindow(#WINDOW_StructureViewer, StructureViewerX, StructureViewerY, StructureViewerWidth, StructureViewerHeight, #PB_Window_Invisible | #PB_Window_SystemMenu | #PB_Window_SizeGadget, Langage(112), WindowID(#WINDOW_Main)) ; "Structure Viewer"

      Top = 6
      CreateGadgetList(WindowID(#WINDOW_StructureViewer))
        StringGadget  (#GADGET_StructureViewer_Name  , 82, 0,   0, 22, "")
        PanelGadget   (#GADGET_StructureViewer_Panel, 6, Top, 0, 22) : Top+22
          AddGadgetItem(#GADGET_StructureViewer_Panel, -1, "Structures")
          AddGadgetItem(#GADGET_StructureViewer_Panel, -1, "Interfaces")
        CloseGadgetList()
        
        ListViewGadget(#GADGET_StructureViewer_List, 6, Top, 0, 0)
         SetGadgetFont (#GADGET_StructureViewer_List, LoadFont(#FONT_StructViewer,EditorFontName, EditorFontSize, EditorFontStyle))

         ButtonGadget  (#GADGET_StructureViewer_InsertName  ,116, 0, 100, 22, Langage(164)) ; "Insert StructureName"
        ButtonGadget  (#GADGET_StructureViewer_InsertStruct,226, 0, 100, 22, Langage(166)) ; "Insert Structure"
        ButtonGadget  (#GADGET_StructureViewer_InsertCopy  ,226, 0, 100, 22, Langage(167)) ; "Insert Copy of Structure"
        ButtonGadget  (#GADGET_StructureViewer_Parent      ,  6, 0,  70, 22, Langage(113)) ; "Parent"
        ; CheckBoxGadget(#GADGET_StructureViewer_OnTop       ,  6, 0, 100, 22, Langage(114)) ; "Stay on top"
        CheckBoxGadget(#GADGET_StructureViewer_InsertType  ,  6, 0, 100, 22, Langage(168)) ; "Include Types"

      ; SetGadgetState(#GADGET_StructureViewer_OnTop , 1) ; TopMost = Default
      ; SetWindowPos_(WindowID(#WINDOW_StructureViewer), #HWND_TOPMOST , 0, 0, 0, 0, #SWP_NOSIZE |#SWP_NOMOVE)

      SetGadgetState(#GADGET_StructureViewer_InsertType,StructureViewerIncludeType)
      DisableGadget(#GADGET_StructureViewer_Parent, 1)

      ResizeStructureViewerWindow()

      If CountList(StructureList$()) = 0 ; only need at first start, as residents doesn't change during the editor life
              
        ; First, load the interfaces..
        ;        
        If SendCompilerRequest(#PB_MSG_Interfaces_Listing)
          If ReadFile(#READFILE_STRUCTURESLISTINGREQUEST, FullPath + "Compilers\PBInterfaces.txt")
          
            ClearList(StructureList$())
            ClearList(InterfaceListBuffer$())
          
            Repeat
              Line$ = ReadString()
              If Line$
                AddSortedStructure(Line$)
              EndIf
            Until Line$ = ""
            CloseFile(#READFILE_STRUCTURESLISTINGREQUEST)
            DeleteFile(FullPath + "Compilers\PBInterfaces.txt")
           
            ; Fill our buffer with the already sorted list
            ;
            ForEach StructureList$()
              AddElement(InterfaceListBuffer$())
              InterfaceListBuffer$() = StructureList$()
            Next
          EndIf
        EndIf

        ; Now load the structure list
        ;        
        If SendCompilerRequest(#PB_MSG_Structures_Listing)
          If ReadFile(#READFILE_STRUCTURESLISTINGREQUEST, FullPath + "Compilers\PBStructures.txt")
           
             ClearList(StructureList$())
             ClearList(StructureListBuffer$())
           
             Repeat
               Line$ = ReadString()
               If Line$
                 AddSortedStructure(UCase(Line$))
               EndIf
             Until Line$ = ""
             CloseFile(#READFILE_STRUCTURESLISTINGREQUEST)
             DeleteFile(FullPath + "Compilers\PBStructures.txt")
             
            ForEach StructureList$()
              AddElement(StructureListBuffer$())
              StructureListBuffer$() = StructureList$()
            Next
          EndIf
        EndIf
      EndIf
      
      StructureViewerMode = #SVM_Structures

      ClearList(StructureList$())
          
      ForEach StructureListBuffer$()
        AddElement(StructureList$())
        StructureList$() = StructureListBuffer$()
      Next
      
      DisplayStructuresList()

      x = IsStructure(CurrentWord$)
      If x <> -1
        DisplayStructureInfo(CurrentWord$, x, 1)
      Else
        DisableGadget(#GADGET_StructureViewer_InsertName, 1)
        ActivateGadget(#GADGET_StructureViewer_Name)
      EndIf
      
      AddKeyboardShortcut(#WINDOW_StructureViewer, #PB_Shortcut_Return   , #GADGET_StructureViewer_Ok)
      AddKeyboardShortcut(#WINDOW_StructureViewer, #PB_Shortcut_Escape   , #GADGET_StructureViewer_Cancel)
      AddKeyboardShortcut(#WINDOW_StructureViewer, #PB_Shortcut_Back     , #GADGET_StructureViewer_Parent)
      
      HideWindow(#WINDOW_StructureViewer, 0)
      
      IsStructureViewerWindow = #True
    EndIf
    
  Else  ; ProcedureBrowser already opened

    x = IsStructure(CurrentWord$)
    If x <> -1
      DisplayStructureInfo(CurrentWord$, x, 1)
    Else
      UseWindow(#WINDOW_StructureViewer)
      ActivateWindow()
    EndIf
  
  EndIf

EndProcedure

Procedure CloseStructureViewerWindow()
  UseWindow(#WINDOW_StructureViewer)
  StructureViewerX      = WindowX()
  StructureViewerY      = WindowY()
  StructureViewerWidth  = WindowWidth()
  StructureViewerHeight = WindowHeight()

  CloseWindow(#WINDOW_StructureViewer)
  UseWindow(#WINDOW_Main)
  
  IsStructureViewerWindow = #False
EndProcedure


Procedure StructureViewerWindowEvents(EventID)

  UseWindow(#WINDOW_StructureViewer)
  
  If EventID = #PB_Event_Menu     ; Little wrapper to map the shortcut events (identified as menu)
    EventID  = #PB_Event_Gadget   ; to normal gadget events...
    GadgetID = EventMenuID()
    If GadgetID = #GADGET_StructureViewer_Parent And GetFocus_() = GadgetID(#GADGET_StructureViewer_Name)
      ;EventID = #PB_Event_Gadget
      GadgetID = #GADGET_StructureViewer_Name
      A$ = GetGadgetText(#GADGET_StructureViewer_Name)
      If A$
        SendMessage_(GadgetID(#GADGET_StructureViewer_Name),#EM_GETSEL,@a,@b)
        If a = b And a <> 0
          SendMessage_(GadgetID(#GADGET_StructureViewer_Name),#EM_SETSEL,a-1,a)
        EndIf
        SendMessage_(GadgetID(#GADGET_StructureViewer_Name),#EM_REPLACESEL,1,"")
      EndIf
    EndIf
    
  Else
    GadgetID = EventGadgetID()
  EndIf

  Select EventID
    Case #PB_Event_CloseWindow

      Quit = 1
      
    Case #PB_Event_Gadget
    
      If GadgetID = #GADGET_StructureViewer_Panel
        DisplayCorrectStructureList()
        
        ActivateGadget(#GADGET_StructureViewer_Name)
        
;      ElseIf GadgetID = #GADGET_StructureViewer_OnTop
;        If GetGadgetState(#GADGET_StructureViewer_OnTop)
;          Flags = #HWND_TOPMOST
;        Else
;          Flags = #HWND_NOTOPMOST            
;        EndIf
;        
;        SetWindowPos_(WindowID(#WINDOW_StructureViewer), Flags, 0, 0, 0, 0, #SWP_NOSIZE |#SWP_NOMOVE)

      ElseIf GadgetID = #GADGET_StructureViewer_InsertType
      
        StructureViewerIncludeType = GetGadgetState(#GADGET_StructureViewer_InsertType)

      ElseIf GadgetID = #GADGET_StructureViewer_Ok Or (GadgetID = #GADGET_StructureViewer_List And EventType() = #PB_EventType_LeftDoubleClick)

        If GetGadgetState(#GADGET_StructureViewer_List) >= 0
          StructureName$ = UCase(GetGadgetText(#GADGET_StructureViewer_List))

          If StructureName$ <> "ENDSTRUCTURE" And StringField(StructureName$,1," ") <> "STRUCTURE"
            If FindString(StructureName$,".",1)
              StructureName$ = GetExtensionPart(StructureName$)
            EndIf

            x = FindString(StructureName$,"[",1) ; xyz.YO[12]
            If x
              StructureName$ = Left(StructureName$,x-1)
            EndIf

            If StructureName$ And StructureName$ <> "B" And StructureName$ <> "W" And StructureName$ <> "L" And StructureName$ <> "F" And StructureName$ <> "S" And Right(StructureName$,1) <> "$"
              x = IsStructure(StructureName$)
              If x <> -1
                If StructureViewerMode = #SVM_Interfaces
                  StructureName$ = GetGadgetText(#GADGET_StructureViewer_List)  ; To have correct case name (not UCase()'d)
                EndIf
              
                DisplayStructureInfo(StructureName$, GetGadgetState(#GADGET_StructureViewer_List), 1)
              Else
                If StructureViewerMode = #SVM_Structures ; Only structures can have built in field which point to another structure
                  x = GetFocus_()
                    MessageRequester("Info", "Structure not found: "+StructureName$, #MB_ICONINFORMATION)
                  SetFocus_(x)
                EndIf
              EndIf
            EndIf
            ;ActivateGadget(#GADGET_StructureViewer_List)
          EndIf
        EndIf
        StructureViewerInSearch = #FALSE
      
      ElseIf GadgetID = #GADGET_StructureViewer_List ;And EventType() = #PB_EventType_LeftClick

        If CountList(StructureHistory()) = 0 Or StructureViewerInSearch = #TRUE
          DisableGadget(#GADGET_StructureViewer_InsertName, 0)
          SetWindowText_(WindowID(#WINDOW_StructureViewer),Langage(112)+"  -  ."+Trim(GetGadgetText(#GADGET_StructureViewer_List)))
          SetGadgetText(#GADGET_StructureViewer_Name, GetGadgetText(#GADGET_StructureViewer_List))
        EndIf
        StructureViewerInSearch = #FALSE
          
      ElseIf GadgetID = #GADGET_StructureViewer_Parent

        If CountList(StructureHistory()) = 0
          SetGadgetState(#GADGET_StructureViewer_List,0)
          SetWindowText_(WindowID(#WINDOW_StructureViewer),Langage(112)+"  -  ."+Trim(GetGadgetText(#GADGET_StructureViewer_List)))
          SetGadgetText(#GADGET_StructureViewer_Name, GetGadgetText(#GADGET_StructureViewer_List))
          ActivateGadget(#GADGET_StructureViewer_List)
        ElseIf CountList(StructureHistory()) > 1
          OldLine = StructureHistory()\Line
          LastElement(StructureHistory()):DeleteElement(StructureHistory())
          DisplayStructureInfo(StructureHistory()\Name$, 0, 0)
          SetGadgetState(#GADGET_StructureViewer_List, OldLine)
        Else
          OldLine = StructureHistory()\Line
          LastElement(StructureHistory()):DeleteElement(StructureHistory())
          DisplayStructuresList()
          DisableGadget(#GADGET_StructureViewer_Parent, 1)
          SetGadgetState(#GADGET_StructureViewer_List,OldLine)
          SetWindowText_(WindowID(#WINDOW_StructureViewer),Langage(112)+"  -  ."+Trim(GetGadgetText(#GADGET_StructureViewer_List)))
          SetGadgetText(#GADGET_StructureViewer_Name, GetGadgetText(#GADGET_StructureViewer_List))
          ActivateGadget(#GADGET_StructureViewer_List)
        EndIf
        StructureViewerInSearch = #FALSE
          
      ElseIf GadgetID =  #GADGET_StructureViewer_InsertName
        
        If CountList(StructureHistory())=0 Or StructureViewerInSearch = #TRUE
          RichEdit_InsertString("."+Trim(GetGadgetText(#GADGET_StructureViewer_List)), #True)
        Else
          RichEdit_InsertString("."+Trim(StructureHistory()\Name$), #True)
        EndIf

      ElseIf GadgetID = #GADGET_StructureViewer_InsertStruct

        If Left(GetGadgetItemText(#GADGET_StructureViewer_List,0,0),10) = "Structure "
          InsertStructure()
        EndIf

      ElseIf GadgetID = #GADGET_StructureViewer_InsertCopy

        If Left(GetGadgetItemText(#GADGET_StructureViewer_List,0,0),10) = "Structure "
          SendMessage_(*RichEditGadget, #EM_REPLACESEL   , 1, "")
          SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 1, 0)
          SendMessage_(*RichEditGadget,#EM_EXGETSEL,0,Range.CHARRANGE)
          Offset = Range\cpMin-SendMessage_(*RichEditGadget,#EM_LINEINDEX,SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,Range\cpMin),0)

          RichEdit_InsertString(GetGadgetItemText(#GADGET_StructureViewer_List,0,0)+Chr(10), #True)

          For a = 1 To CountGadgetItems(#GADGET_StructureViewer_List)-1
            RichEdit_InsertString(Space(Offset)+GetGadgetItemText(#GADGET_StructureViewer_List,a,0)+Chr(10), #True)
          Next a

        EndIf

      ElseIf GadgetID = #GADGET_StructureViewer_Name

        If EventType() = #PB_EventType_Change
          StructureViewerInSearch = #TRUE
          SearchStructure(GetGadgetText(#GADGET_StructureViewer_Name))
        EndIf

      ElseIf GadgetID = #GADGET_StructureViewer_Cancel

        Quit = 1

      EndIf
    
  EndSelect
  
  If Quit
    CloseStructureViewerWindow()
  EndIf
  
EndProcedure

; Test line...
; 
;LoadFont(0, "Courier", 10) : OpenWindow(#WINDOW_Main, 100, 100, 800, 400, #PB_Window_SystemMenu, "") : OpenStructureViewerWindow() : Repeat : Event = WaitWindowEvent() : StructureViewerWindowEvents(Event) : Until Event = #PB_Event_CloseWindow
; ExecutableFormat=Windows
; UseMainFile=PureBasic IDE.pb
; EOF